﻿// --------------------------------
// See LICENCE file at project root
// File : algorithm/omp/upward.hpp
// --------------------------------
#ifndef SCALFMM_ALGORITHMS_MPI_UPWARD_HPP
#define SCALFMM_ALGORITHMS_MPI_UPWARD_HPP

#ifdef _OPENMP

#include <omp.h>

#include "scalfmm/algorithms/omp/upward.hpp"
// #include "scalfmm/operators/m2m.hpp"
// #include "scalfmm/operators/tags.hpp"
// #include "scalfmm/tree/utils.hpp"
// #include "scalfmm/utils/massert.hpp"
// #include "scalfmm/utils/math.hpp"

#include <cpp_tools/parallel_manager/parallel_manager.hpp>

namespace scalfmm::algorithms::mpi::pass
{
    /// @brief Construct the vector of dependencies (child group)
    /// @tparam IteratorType
    /// @tparam MortonType
    /// @tparam Dependencies_t
    /// @tparam dimension
    /// @param begin first iterator on the groups of cells (child)
    /// @param end  last iterator on the groups of cells (child)
    /// @param parent_morton_index the parent index
    /// @param dependencies  the vector of dependencies
    template<int dimension, typename IteratorType, typename MortonType, typename Dependencies_t>
    void build_dependencies(IteratorType begin, IteratorType end, MortonType const& parent_morton_index,
                            Dependencies_t& dependencies)
    {
        for(auto grp_ptr = begin; grp_ptr != end; ++grp_ptr)
        {
            auto const& csymb = (*grp_ptr)->csymbolics();
            // iterate on the cells in the same group
            // we move forward in the index vector
            if(parent_morton_index == (csymb.starting_index >> dimension))
            {
                // std::cout << " add depend for grp with Int [" << csymb.starting_index << ", " << csymb.ending_index
                //           << "]" << std::endl;
                dependencies.push_back(&(grp_ptr->get()->ccomponent(0).cmultipoles(0)));
            }
            else
            {
                break;
            }
        }
    }
    /**
     * @brief Perform the communications for the children level
     *
     * @tparam Tree
     * @tparam Approximation
     */
    template<typename Tree>
    inline auto start_communications(const int& level, Tree& tree) -> void
    {
        using value_type = typename Tree::base_type::cell_type::value_type;
        using dep_type = typename Tree::group_of_cell_type::symbolics_type::ptr_multi_dependency_type;

        static constexpr std::size_t dimension = Tree::base_type::box_type::dimension;
        static constexpr int nb_inputs = Tree::cell_type::storage_type::inputs_size;
        //
        // number of theoretical children
        constexpr int nb_children = math::pow(2, dimension);
        static constexpr auto prio{omp::priorities::max};
        //
        auto& para = tree.get_parallel_manager();
        auto comm = para.get_communicator();
        auto rank = comm.rank();
        int nb_proc = comm.size();
        int tag_nb = 1200 + 10 * level;
        int tag_data = 1201 + 10 * level;
        auto mpi_int_type = cpp_tools::parallel_manager::mpi::get_datatype<int>();

        auto level_child = level + 1;
        // get size of multipoles
        auto it_group_child = tree.begin_mine_cells(level_child);
        auto it_cell_child = it_group_child->get()->begin();
        auto const& m1 = it_cell_child->cmultipoles();
        int size{int(nb_inputs * m1.at(0).size()) * nb_children};

        std::vector<dep_type> deps, dependencies;
        auto const& distrib = tree.get_cell_distribution(level);

        if(rank > 0)
        {
            // If we send the multipoles, they must have been updated by the M2M of the previous level!
            // Check if you have to send something on the left (rank-1)
            // We are at level l and the children are at level l+1.
            // The ghost at the child level(on the right)
            //   must be updated if the parent index of the first child index is not equal to(lower than)
            //     the first index at level l.
            int count{0};
            // serialization
            std::vector<value_type> buffer(size, 0);
            // check if we have some child to send (if there exists they starts at the first group)
            auto first_group_of_child = tree.begin_mine_cells(level_child);
            auto first_index_child = first_group_of_child->get()->component(0).index();
            auto parent_of_first_index_child = first_index_child >> dimension;
            //
            // We are at the parent level where we apply the m2m operator (same number as the in the distribution)
            auto first_group = tree.begin_mine_cells(level);
            auto first_index_parent = first_group->get()->component(0).index();
            // std::cout << "upward first_index_parent " << first_index_parent << " distrib " << distrib[rank][0]
            //           << std::endl
            //           << std::flush;
            auto dep_parent = &(first_group->get()->component(0).clocals(0));
            // &(group_ptr->ccomponent(0).clocals(0));
            // Check if I have the parent
            // std::cout << " upward " << first_index_parent << "  parent of last child index "
            //           << parent_of_first_index_child << "  comm " << std::boolalpha
            //           << (first_index_parent > parent_of_first_index_child) << std::endl
            //           << std::flush;
            // std::cout << "upward send comm to " << rank - 1 << std::endl << std::flush;
            auto parent_morton_index = distrib[rank - 1][1] - 1;
            // std::cout << "upward  parent_morton_index " << parent_morton_index << std::endl << std::flush;

            build_dependencies<dimension>(first_group_of_child, tree.end_mine_cells(level_child), parent_morton_index,
                                          dependencies);
            // io::print("upward(send) dependencies(inout) ", dependencies);
            // std::cout << "upward(send) dep_parent(in) " << dep_parent << std::endl << std::flush;
            std::cout << std::flush;
            // dep parent ?????? depend(in : dep_parent)

#pragma omp task firstprivate(rank, tag_data, tag_nb, mpi_int_type, size, dimension)                                   \
  depend(iterator(std::size_t it = 0 dependencies.size()), in : (dependencies[it])[0]),                                \
  priority(prio)
            {
                auto index = first_index_parent;
                // std::cout << "upward(task) index " << index << "  parent_of_first_index_child "
                //           << parent_of_first_index_child << std::endl
                //           << std::flush;
                if(index > parent_of_first_index_child)
                {
                    // index is now the parent of the first child
                    index = first_index_child >> dimension;
                    // I have to send
                    // find the number of children to send (get pointer on multipoles !!)
                    // Construct an MPI datatype

                    // serialization
                    auto it_group = tree.begin_mine_cells(level_child);
                    auto it_cell = first_group_of_child->get()->begin();
                    auto next = scalfmm::component::generate_linear_iterator(1, it_group, it_cell);

                    // std::vector<> cells_to_send();
                    //  We construct an MPI DATA_TYPE
                    // MPI_Datatype mult_data_type;
                    auto it = std::begin(buffer);
                    for(int i = 0; i < nb_children - 1; ++i, next())
                    {
                        if(index < (it_cell->index() >> dimension))
                        {
                            break;
                        }
                        // std::cout << "upward(task) Check children  P " << index << " C " << it_cell->index()
                        //           << " level " << it_cell->csymbolics().level << std::endl
                        //           << std::flush;
                        // copy the multipoles in the buffer
                        auto const& m = it_cell->cmultipoles();

                        auto nb_m = m.size();

                        for(std::size_t i{0}; i < nb_m; ++i)
                        {
                            auto const& ten = m.at(i);
                            std::copy(std::begin(ten), std::end(ten), it);
                            it += ten.size();
                        }
                        ++count;
                    }
                }
                // std::cout << "upward(task) nb_send = " << count << std::endl;

                comm.isend(&count, 1, mpi_int_type, rank - 1, tag_nb);

                if(count != 0)
                {
                    // loop to serialize the multipoles
                    auto mpi_type = cpp_tools::parallel_manager::mpi::get_datatype<value_type>();

                    comm.isend(buffer.data(), size, mpi_type, rank - 1, tag_data);
                    // std::cout << "upward(task) buffer:";
                    // for(int i = 0; i < buffer.size(); ++i)
                    // {
                    //     std::cout << "  " << buffer[i];
                    // }
                    // std::cout << " \n" << std::flush;
                }
            }
            // std::cout << "  end submitted task send \n" << std::flush;
        }

        if(rank == nb_proc - 1)
        {
            return;
        }
        // Add task dependencies
        //
        // We receive the cells (at most 2^d - 1) to have all the children of
        //  the last father cell I own. These cells go into the first ghost group on the right.
        // dep(out) these cells
        // dep(out) group_parent_dep[0]??
        // std::cout << " start receive \n" << std::flush;

        deps.clear();
        dependencies.clear();
        // last group
        // auto it_group = --(tree.end_mine_cells(level_child));
        it_group_child = --(tree.end_mine_cells(level_child));
        // std::cout << "it_group ok " << std::endl << std::flush;

        auto gs = it_group_child->get()->size();
        // std::cout << " gs = " << gs << std::endl << std::flush;
        int nb_grp_dep = std::min(static_cast<int>(nb_children / gs + 1),
                                  static_cast<int>(std::distance(it_group_child, tree.end_cells(level_child))));
        // std::cout << " nb_grp_dep = " << nb_grp_dep << " static_cast<int>(nb_children / gs + 1) "
        //           << static_cast<int>(nb_children / gs + 1) << std::endl
        //           << " all " << static_cast<int>(std::distance(it_group, tree.end_cells(level_child))) << std::endl
        //           << std::flush;

        auto it_group_parent = --(tree.end_mine_cells(level));
        auto it_grp = it_group_child;
        for(int i = 0; i < nb_grp_dep; ++i, ++it_grp)
        {
            deps.push_back(&(it_grp->get()->ccomponent(0).cmultipoles(0)));
        }
        // io::print("upward receive dependencies(M2M) ", deps);
        auto group_parent_dep = it_group_parent->get()->ccomponent(0).cmultipoles(0);
        // #pragma omp task shared(rank, mpi_int_type, size, tag_data, tag_nb)                                                    \
//   depend(iterator(std::size_t it = 0 : deps.size()), out : (deps[it])[0]) priority(prio)
        {
            int count{-1};
            comm.recv(&count, 1, mpi_int_type, rank + 1, tag_nb);
            // std::cout << "upward()     We receive " << count << " cell(s)\n" << std::flush;
            // use a recv
            if(count > 0)
            {
                std::vector<value_type> buffer(size);
                auto mpi_type = cpp_tools::parallel_manager::mpi::get_datatype<value_type>();
                comm.recv(buffer.data(), size, mpi_type, rank + 1, tag_data);
                //
                // std::cout << "upward()     buffer:";
                // for(int i = 0; i < buffer.size(); ++i)
                // {
                //     std::cout << "  " << buffer[i];
                // }
                std::cout << std::endl << std::flush;
                // first right ghost
                auto it_group = tree.end_mine_cells(level_child);
                auto it_cell = it_group->get()->begin();
                // linear iterator on cells
                auto next1 = scalfmm::component::generate_linear_iterator(1, it_group, it_cell);
                auto it = std::begin(buffer);

                for(int i = 0; i < count; ++i, next1())
                {
                    // copy the multipoles in the buffer
                    auto& m = it_cell->multipoles();

                    auto nb_m = m.size();
                    // std::cout << "upward() cell index: " << it_cell->index() << " level " << it_cell->csymbolics().level
                    //           << "\n"
                    //           << std::flush;
                    for(std::size_t i{0}; i < nb_m; ++i)
                    {
                        auto& ten = m.at(i);
                        // std::cout << "upward()    ten(before) " << ten << std::endl << std::flush;
                        std::copy(it, it + ten.size(), std::begin(ten));
                        it += ten.size();
                        // std::cout << "upward()    ten(after) " << ten << std::endl << std::flush;
                    }
                }
            }
        }
        // std::cout << " END start_communications" << std::endl << std::flush;
        // #pragma omp taskwait
        //
    }

    /// @brief This function constructs the local approximation for all the cells of the tree by applying the operator
    /// m2m
    ///
    /// @param tree   the tree target
    /// @param approximation the approximation to construct the local approximation
    ///
    template<typename Tree, typename Approximation>
    inline auto upward(Tree& tree, Approximation const& approximation) -> void
    {
        auto leaf_level = tree.height() - 1;

        // upper working level is
        const int top_height = tree.box().is_periodic() ? 0 : 2;
        // const int start_duplicated_level = tree.start_duplicated_level();
        //
        // int top = start_duplicated_level < 0 ? top_height : start_duplicated_level - 1;
        int top = top_height;
        for(int level = leaf_level - 1; level >= top /*top_height*/; --level)   // int because top_height could be 0
        {
            // std::cout << "M2M : " << level + 1 << " -> " << level << std::endl << std::flush;
            //
            start_communications(level, tree);
            // std::cout << "  end comm " << std::endl << std::flush;

            omp::pass::upward_level(level, tree, approximation);
            // std::cout << "  end upward_level " << level << std::endl << std::flush;
        }
        // std::cout << "end upward " << std::endl << std::flush;

        //

        // for(int level = start_duplicated_level; level >= top_height; --level)   // int because top_height could be 0
        // {
        //     std::cout << "Level duplicated (seq): " << level << std::endl;
        //     upward_level(level, tree, approximation);
        // }
    }
}   // namespace scalfmm::algorithms::mpi::pass

#endif   // _OPENMP
#endif   // SCALFMM_ALGORITHMS_MPI_UPWARD_HPP
